package org.blankapp.flutterplugins.flutter_photo_picker;

import android.Manifest;
import android.app.Activity;
import android.content.ContentResolver;
import android.content.Context;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.database.Cursor;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Matrix;
import android.graphics.Point;
import android.media.ExifInterface;
import android.media.MediaMetadataRetriever;
import android.media.ThumbnailUtils;
import android.net.Uri;
import android.os.Build;
import android.os.Environment;
import android.provider.MediaStore;
import android.text.TextUtils;
import android.webkit.MimeTypeMap;

import com.zhihu.matisse.internal.utils.PhotoMetadataUtils;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Arrays;

class FlutterPhotoPickerUtils {
    public static String getRealFilePath(Context context, Uri uri) {
        if (null == uri) return null;
        final String scheme = uri.getScheme();
        String realPath = null;
        if (scheme == null)
            realPath = uri.getPath();
        else if (ContentResolver.SCHEME_FILE.equals(scheme)) {
            realPath = uri.getPath();
        } else if (ContentResolver.SCHEME_CONTENT.equals(scheme)) {
            Cursor cursor = context.getContentResolver().query(uri,
                    new String[]{MediaStore.Images.ImageColumns.DATA},
                    null, null, null);
            if (null != cursor) {
                if (cursor.moveToFirst()) {
                    int index = cursor.getColumnIndex(MediaStore.Images.ImageColumns.DATA);
                    if (index > -1) {
                        realPath = cursor.getString(index);
                    }
                }
                cursor.close();
            }
        }
        if (TextUtils.isEmpty(realPath)) {
            if (uri != null) {
                String uriString = uri.toString();
                int index = uriString.lastIndexOf("/");
                String imageName = uriString.substring(index);
                File storageDir;

                storageDir = Environment.getExternalStoragePublicDirectory(
                        Environment.DIRECTORY_PICTURES);
                File file = new File(storageDir, imageName);
                if (file.exists()) {
                    realPath = file.getAbsolutePath();
                } else {
                    storageDir = context.getExternalFilesDir(Environment.DIRECTORY_PICTURES);
                    File file1 = new File(storageDir, imageName);
                    realPath = file1.getAbsolutePath();
                }
            }
        }
        return realPath;
    }

    public static String getMimeType(Context context, Uri uri) {
        String mimeType = null;
        if (uri.getScheme().equals(ContentResolver.SCHEME_CONTENT)) {
            ContentResolver cr = context.getContentResolver();
            mimeType = cr.getType(uri);
        } else {
            String fileExtension = MimeTypeMap.getFileExtensionFromUrl(uri.toString());
            mimeType = MimeTypeMap.getSingleton().getMimeTypeFromExtension(fileExtension.toLowerCase());
        }
        return mimeType;
    }

    public static Point getImageSize(Context context, Uri uri) {
        ContentResolver contentProvider = context.getContentResolver();
        Point imageSize = PhotoMetadataUtils.getBitmapBound(contentProvider, uri);
        int w = imageSize.x;
        int h = imageSize.y;

        return new Point(w, h);
    }

    public static Point getVideoSize(Context context, Uri uri) {
        ContentResolver contentProvider = context.getContentResolver();
        MediaMetadataRetriever retriever = new MediaMetadataRetriever();
        retriever.setDataSource(PhotoMetadataUtils.getPath(contentProvider, uri));
        int w = Integer.valueOf(retriever.extractMetadata(MediaMetadataRetriever.METADATA_KEY_VIDEO_WIDTH));
        int h = Integer.valueOf(retriever.extractMetadata(MediaMetadataRetriever.METADATA_KEY_VIDEO_HEIGHT));
        retriever.release();

        return new Point(w, h);
    }

    public static File createThumbnail(Context context, String originalImagePath, int maxWidth, int maxHeight, int quality) throws IOException {
        String mimeType = getMimeType(context, Uri.fromFile(new File(originalImagePath)));
        Bitmap original;

        if (mimeType.startsWith("video")) {
            original = ThumbnailUtils.createVideoThumbnail(originalImagePath, MediaStore.Images.Thumbnails.MINI_KIND);
        } else {
            original = BitmapFactory.decodeFile(originalImagePath);
        }

        int width = original.getWidth();
        int height = original.getHeight();

        ExifInterface originalExif = new ExifInterface(originalImagePath);
        int originalOrientation = originalExif.getAttributeInt(ExifInterface.TAG_ORIENTATION, 1);

        Matrix rotationMatrix = new Matrix();
        int rotationAngleInDegrees = getRotationInDegreesForOrientationTag(originalOrientation);
        rotationMatrix.postRotate(rotationAngleInDegrees);

        float ratioBitmap = (float) width / (float) height;
        float ratioMax = (float) maxWidth / (float) maxHeight;

        int finalWidth = maxWidth;
        int finalHeight = maxHeight;

        if (ratioMax > 1) {
            finalWidth = (int) ((float) maxHeight * ratioBitmap);
        } else {
            finalHeight = (int) ((float) maxWidth / ratioBitmap);
        }

        Bitmap resized = Bitmap.createScaledBitmap(original, finalWidth, finalHeight, true);
        resized = Bitmap.createBitmap(resized, 0, 0, finalWidth, finalHeight, rotationMatrix, true);

        File imageDirectory = new File(getTmpDir(context));

        if (!imageDirectory.exists()) {
            imageDirectory.mkdirs();
        }

        File originalImageFile = new File(originalImagePath);
        File resizeImageFile = new File(imageDirectory, originalImageFile.getName() + "-thumbnail.jpg");

        OutputStream os = new BufferedOutputStream(new FileOutputStream(resizeImageFile));
        resized.compress(Bitmap.CompressFormat.JPEG, quality, os);

        os.close();
        original.recycle();
        resized.recycle();

        return resizeImageFile;
    }

    /** returns true, if permission present in manifest, otherwise false */
    private static boolean isPermissionPresentInManifest(Context context, String permissionName) {
        try {
            PackageManager packageManager = context.getPackageManager();
            PackageInfo packageInfo =
                    packageManager.getPackageInfo(context.getPackageName(), PackageManager.GET_PERMISSIONS);

            String[] requestedPermissions = packageInfo.requestedPermissions;
            return Arrays.asList(requestedPermissions).contains(permissionName);
        } catch (PackageManager.NameNotFoundException e) {
            e.printStackTrace();
            return false;
        }
    }

    /**
     * Camera permission need request if it present in manifest, because for M or great for take Photo
     * ar Video by intent need it permission, even if the camera permission is not used.
     *
     * <p>Camera permission may be used in another package, as example flutter_barcode_reader.
     * https://github.com/flutter/flutter/issues/29837
     *
     * @return returns true, if need request camera permission, otherwise false
     */
    static boolean needRequestCameraPermission(Context context) {
        boolean greatOrEqualM = Build.VERSION.SDK_INT >= Build.VERSION_CODES.M;
        return greatOrEqualM && isPermissionPresentInManifest(context, Manifest.permission.CAMERA);
    }

    private static int getRotationInDegreesForOrientationTag(int orientationTag) {
        switch (orientationTag) {
            case ExifInterface.ORIENTATION_ROTATE_90:
                return 90;
            case ExifInterface.ORIENTATION_ROTATE_270:
                return -90;
            case ExifInterface.ORIENTATION_ROTATE_180:
                return 180;
            default:
                return 0;
        }
    }

    private static String getTmpDir(Context context) {
        String tmpDir = context.getCacheDir() + "/flutter_photo_picker";
        new File(tmpDir).mkdir();

        return tmpDir;
    }
}
